import pyarrow
import dataclasses as dc

def _repr(x):
	if isinstance(x, pyarrow.Array):
		s = ', '.join(repr(v.as_py()) for v in x[:3])
		if len(x) > 3:
			s += ', ...'
		return f'[{s}]'
	elif isinstance(x, tuple):
		s = ', '.join(_repr(v) for v in x)
		return f'({s})'
	elif dc.is_dataclass(x):
		s = ', '.join(f'{f.name}={_repr(getattr(x, f.name))}' for f in dc.fields(type(x)))
		return f'{type(x).__name__}({s})'
	else:
		return repr(x)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os, sys

project = 'peppi-py'
copyright = '2024, melkor'
author = 'melkor'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.linkcode',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

language = 'en'

sys.path.insert(0, os.path.abspath('..'))

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']

def linkcode_resolve(domain, info):
    if domain != 'py':
        return None
    if not (module := info['module']):
        return None
    filename = module.replace('.', '/')
    if module == 'peppi_py':
        filename += '/__init__'
    return f'https://github.com/hohav/peppi-py/blob/main/{filename}.py'



from dataclasses import dataclass
from pyarrow.lib import Int8Array, Int16Array, Int32Array, Int64Array, UInt8Array, UInt16Array, UInt32Array, UInt64Array, FloatArray, DoubleArray
from .util import _repr

@dataclass(slots=True)
class End:
	__repr__ = _repr
	latest_finalized_frame: Int32Array | None = None

@dataclass(slots=True)
class Position:
	__repr__ = _repr
	x: FloatArray
	y: FloatArray

@dataclass(slots=True)
class Start:
	__repr__ = _repr
	random_seed: UInt32Array
	scene_frame_counter: UInt32Array | None = None

@dataclass(slots=True)
class TriggersPhysical:
	__repr__ = _repr
	l: FloatArray
	r: FloatArray

@dataclass(slots=True)
class Velocities:
	__repr__ = _repr
	self_x_air: FloatArray
	self_y: FloatArray
	knockback_x: FloatArray
	knockback_y: FloatArray
	self_x_ground: FloatArray

@dataclass(slots=True)
class Velocity:
	__repr__ = _repr
	x: FloatArray
	y: FloatArray

@dataclass(slots=True)
class Item:
	__repr__ = _repr
	type: UInt16Array
	state: UInt8Array
	direction: FloatArray
	velocity: Velocity
	position: Position
	damage: UInt16Array
	timer: FloatArray
	id: UInt32Array
	misc: tuple[UInt8Array, UInt8Array, UInt8Array, UInt8Array] | None = None
	owner: Int8Array | None = None

@dataclass(slots=True)
class Post:
	__repr__ = _repr
	character: UInt8Array
	state: UInt16Array
	position: Position
	direction: FloatArray
	percent: FloatArray
	shield: FloatArray
	last_attack_landed: UInt8Array
	combo_count: UInt8Array
	last_hit_by: UInt8Array
	stocks: UInt8Array
	state_age: FloatArray | None = None
	state_flags: tuple[UInt8Array, UInt8Array, UInt8Array, UInt8Array, UInt8Array] | None = None
	misc_as: FloatArray | None = None
	airborne: UInt8Array | None = None
	ground: UInt16Array | None = None
	jumps: UInt8Array | None = None
	l_cancel: UInt8Array | None = None
	hurtbox_state: UInt8Array | None = None
	velocities: Velocities | None = None
	hitlag: FloatArray | None = None
	animation_index: UInt32Array | None = None

@dataclass(slots=True)
class Pre:
	__repr__ = _repr
	random_seed: UInt32Array
	state: UInt16Array
	position: Position
	direction: FloatArray
	joystick: Position
	cstick: Position
	triggers: FloatArray
	buttons: UInt32Array
	buttons_physical: UInt16Array
	triggers_physical: TriggersPhysical
	raw_analog_x: Int8Array | None = None
	percent: FloatArray | None = None
	raw_analog_y: Int8Array | None = None

@dataclass(slots=True)
class Data:
	__repr__ = _repr
	pre: Pre
	post: Post

@dataclass(slots=True)
class PortData:
	__repr__ = _repr
	leader: Data
	follower: Data | None = None

@dataclass(slots=True)
class Frame:
	__repr__ = _repr
	id: object
	ports: tuple[PortData]



from dataclasses import dataclass
from enum import IntEnum
from .frame import Frame
from .util import _repr

class Port(IntEnum):
	P1 = 0
	P2 = 1
	P3 = 2
	P4 = 3

class PlayerType(IntEnum):
	HUMAN = 0
	CPU = 1
	DEMO = 2

class Language(IntEnum):
	JAPANESE = 0
	ENGLISH = 1

class DashBack(IntEnum):
	UCF = 1
	ARDUINO = 2

class ShieldDrop(IntEnum):
	UCF = 1
	ARDUINO = 2

class EndMethod(IntEnum):
	UNRESOLVED = 0
	TIME = 1
	GAME = 2
	RESOLVED = 3
	NO_CONTEST = 7

@dataclass(slots=True)
class Scene:
	__repr__ = _repr
	major: int
	minor: int

@dataclass(slots=True)
class Match:
	__repr__ = _repr
	id: str
	game: int
	tiebreaker: int

@dataclass(slots=True)
class Slippi:
	__repr__ = _repr
	version: tuple[int, int, int]

@dataclass(slots=True)
class Netplay:
	__repr__ = _repr
	name: str
	code: str
	suid: str | None = None

@dataclass(slots=True)
class Team:
	__repr__ = _repr
	color: int
	shade: int

@dataclass(slots=True)
class Ucf:
	__repr__ = _repr
	dash_back: DashBack | None
	shield_drop: ShieldDrop | None

@dataclass(slots=True)
class Player:
	__repr__ = _repr
	port: Port
	character: int
	type: PlayerType
	stocks: int
	costume: int
	team: Team | None
	handicap: int
	bitfield: int
	cpu_level: int | None
	offense_ratio: float
	defense_ratio: float
	model_scale: float
	ucf: Ucf | None = None
	name_tag: str | None = None
	netplay: Netplay | None = None

@dataclass(slots=True)
class Start:
	__repr__ = _repr
	slippi: Slippi
	bitfield: tuple[int, int, int, int]
	is_raining_bombs: bool
	is_teams: bool
	item_spawn_frequency: int
	self_destruct_score: int
	stage: int
	timer: int
	item_spawn_bitfield: tuple[int, int, int, int, int]
	damage_ratio: float
	players: tuple[Player, ...]
	random_seed: int
	is_pal: bool | None = None
	is_frozen_ps: bool | None = None
	scene: Scene | None = None
	language: Language | None = None
	match: Match | None = None

@dataclass(slots=True)
class PlayerEnd:
	__repr__ = _repr
	port: Port
	placement: int

@dataclass(slots=True)
class End:
	__repr__ = _repr
	method: EndMethod
	lras_initiator: Port | None = None
	players: tuple[PlayerEnd, ...] | None = None

@dataclass(slots=True)
class Game:
	__repr__ = _repr
	start: Start
	end: End
	metadata: dict
	frames: Frame | None



from math import isclose
from pathlib import Path
from peppi_py import read_slippi, read_peppi
from peppi_py.game import *

def test_basic_game():
	game = read_slippi(Path(__file__).parent.joinpath('data/game.slp').as_posix())

	assert game.metadata == {
		'startAt': '2018-06-22T07:52:59Z',
		'lastFrame': 5085,
		'playedOn': 'dolphin',
		'players': {
			'0': {'characters': {'18': 5209}}, # Marth
			'1': {'characters': {'1': 5209}}, # Fox
		},
	}

	assert game.start == Start(
		slippi=Slippi(
			version=(1, 0, 0),
		),
		bitfield=(50, 1, 134, 76),
		is_raining_bombs=False,
		is_teams=False,
		item_spawn_frequency=-1,
		self_destruct_score=-1,
		stage=8, # Yoshi's Story
		timer=480,
		item_spawn_bitfield=(255, 255, 255, 255, 255),
		damage_ratio=1.0,
		players=(
			Player(
				port=Port.P1,
				character=9, # Marth
				type=PlayerType.HUMAN,
				stocks=4,
				costume=3,
				team=None,
				handicap=9,
				bitfield=192,
				cpu_level=None,
				offense_ratio=1.0,
				defense_ratio=1.0,
				model_scale=1.0,
				ucf=Ucf(
					dash_back=None,
					shield_drop=None,
				),
				name_tag=None,
				netplay=None,
			),
			Player(
				port=Port.P2,
				character=2, # Fox
				type=PlayerType.CPU,
				stocks=4,
				costume=0,
				team=None,
				handicap=9,
				bitfield=64,
				cpu_level=1,
				offense_ratio=1.0,
				defense_ratio=1.0,
				model_scale=1.0,
				ucf=Ucf(
					dash_back=None,
					shield_drop=None,
				),
				name_tag=None,
				netplay=None,
			),
		),
		random_seed=3803194226,
		is_pal=None,
		is_frozen_ps=None,
		scene=None,
		language=None,
		match=None,
	)

	assert game.end == End(
		method=EndMethod.RESOLVED,
		lras_initiator=None,
		players=None,
	)

	assert len(game.frames.id) == 5209
	p1 = game.frames.ports[0].leader.pre
	p2 = game.frames.ports[1].leader.pre
	assert p1.position.x[1000].as_py() == 56.818748474121094
	assert p1.position.y[1000].as_py() == -18.6373291015625
	assert p2.position.x[1000].as_py() == 42.195167541503906
	assert p2.position.y[1000].as_py() == 9.287015914916992



#!/usr/bin/env python
from dataclasses import dataclass
from ._peppi import read_peppi as _read_peppi, read_slippi as _read_slippi
from .game import End, Game, Start
from .parse import dc_from_json, frames_from_sa

def read_peppi(path: str, skip_frames: bool=False) -> Game:
	g = _read_peppi(path, skip_frames)
	return Game(dc_from_json(Start, g.start), dc_from_json(End, g.end), g.metadata, frames_from_sa(g.frames))

def read_slippi(path: str, skip_frames: bool=False) -> Game:
	g = _read_slippi(path, skip_frames)
	return Game(dc_from_json(Start, g.start), dc_from_json(End, g.end), g.metadata, frames_from_sa(g.frames))



import sys, types, typing
import pyarrow
import dataclasses as dc
from inflection import underscore
from enum import Enum
from .frame import Data, Frame, PortData

def _repr(x):
	if isinstance(x, pyarrow.Array):
		s = ', '.join(repr(v.as_py()) for v in x[:3])
		if len(x) > 3:
			s += ', ...'
		return f'[{s}]'
	elif isinstance(x, tuple):
		s = ', '.join(_repr(v) for v in x)
		return f'({s})'
	elif dc.is_dataclass(x):
		s = ', '.join(f'{f.name}={_repr(getattr(x, f.name))}' for f in dc.fields(type(x)))
		return f'{type(x).__name__}({s})'
	else:
		return repr(x)

def unwrap_union(cls):
	if typing.get_origin(cls) is types.UnionType:
		return typing.get_args(cls)[0]
	else:
		return cls

def field_from_sa(cls, arr):
	if arr is None:
		return None
	cls = unwrap_union(cls)
	if dc.is_dataclass(cls):
		return dc_from_sa(cls, arr)
	elif typing.get_origin(cls) is tuple:
		return tuple_from_sa(cls, arr)
	else:
		return arr

def arr_field(arr, dc_field):
	try:
		return arr.field(dc_field.name)
	except KeyError:
		if dc_field.default is dc.MISSING:
			raise
		else:
			return dc_field.default

def dc_from_sa(cls, arr):
	return cls(*(field_from_sa(f.type, arr_field(arr, f)) for f in dc.fields(cls)))

def tuple_from_sa(cls, arr):
	return tuple((field_from_sa(t, arr.field(str(idx))) for (idx, t) in enumerate(typing.get_args(cls))))

def frames_from_sa(arrow_frames):
	if arrow_frames is None:
		return None
	ports = []
	port_arrays = arrow_frames.field('ports')
	for p in ('P1', 'P2', 'P3', 'P4'):
		try: port = port_arrays.field(p)
		except KeyError: continue
		leader = dc_from_sa(Data, port.field('leader'))
		try: follower = dc_from_sa(Data, port.field('follower'))
		except KeyError: follower = None
		ports.append(PortData(leader, follower))
	return Frame(arrow_frames.field('id'), tuple(ports))

def field_from_json(cls, json):
	if json is None:
		return None
	cls = unwrap_union(cls)
	if dc.is_dataclass(cls):
		return dc_from_json(cls, json)
	elif typing.get_origin(cls) is tuple:
		return tuple_from_json(cls, json)
	elif issubclass(cls, Enum):
		return cls[underscore(json).upper()]
	else:
		return json

def dc_from_json(cls, json):
	return cls(*(field_from_json(f.type, json.get(f.name)) for f in dc.fields(cls)))

def tuple_from_json(cls, json):
	child_cls = typing.get_args(cls)[0]

	if type(json) is dict:
		items = json.items()
	else:
		items = enumerate(json)

	return tuple((field_from_json(child_cls, val) for (idx, val) in items))
